# -*- coding: utf-8 -*-
import abc
from typing import Tuple, Optional

import torch
from PIL import Image
from PIL.JpegImagePlugin import JpegImageFile
from torch.utils.data import Dataset
from torchvision.transforms import Compose


class ObjectDetectionDataset(Dataset, metaclass=abc.ABCMeta):
    """
    Abstract Dataset class for object detection
    """

    def __init__(self, transform: Optional[Compose] = None):
        self._transform = transform

    def _open_img(self, img_path: str):
        img = Image.open(img_path)
        if self._transform:
            img = self._transform(img)
        return img

    @abc.abstractmethod
    def __getitem__(self, index: int) -> Tuple[JpegImageFile, torch.Tensor]:
        """
        Get image and labels
        Args:
            index: index of data

        Returns:
            jpg image and labels
        """
        pass

    @abc.abstractmethod
    def __len__(self) -> int:
        pass



# -*- coding: utf-8 -*-
from nn_torch.cv.models.detection.utils import calc_iou, BBoxFmt


def test_iou():
    iou = calc_iou([4, 5, 7, 7], [5, 6, 10, 10], fmt=BBoxFmt.X_MIN_Y_MIN_X_MAX_Y_MAX)
    assert abs(iou - 1 / 12) <= 1e-6

    iou = calc_iou([1, 1, 2, 2], [10, 10, 11, 11], fmt=BBoxFmt.X_MIN_Y_MIN_X_MAX_Y_MAX)
    assert iou <= 1e-6



# -*- coding: utf-8 -*-
import abc
from typing import List

import torch
import torch.nn as nn
import torch.nn.functional as F

from nn_torch.cv.models.mobilenet import MobileNetV3Large
from nn_torch.cv.models.resnet import ResNet34


class FPN(nn.Module):
    """
    FPN Module

    References: http://arxiv.org/abs/1612.03144
    """

    class Backbone(nn.Module, metaclass=abc.ABCMeta):
        """Backbone for FPN"""

        @property
        @abc.abstractmethod
        def in_channels_tuple(self) -> tuple:
            pass

        @abc.abstractmethod
        def bottom_up(self, x: torch.Tensor):
            pass

    def __init__(self, body: Backbone, *, features_num: int, out_channels: int):
        super().__init__()
        self._features_num = features_num
        self._out_channels = out_channels

        self.body = body

        self.inner_block_modules = nn.ModuleList(
            [nn.Conv2d(kernel_size=1, in_channels=i, out_channels=out_channels) for i in
             body.in_channels_tuple]
        )
        self.layer_block_modules = nn.ModuleList([
            nn.Conv2d(kernel_size=3, in_channels=out_channels, out_channels=out_channels, padding=1) for _ in
            range(len(body.in_channels_tuple))]
        )

    @property
    def features_num(self) -> int:
        """
        Number of feature map
        """
        return self._features_num

    @property
    def out_channels(self) -> int:
        return self._out_channels

    def forward(self, x: torch.Tensor) -> List[torch.Tensor]:
        layers = self.body.bottom_up(x)

        last_layer = self.inner_block_modules[0](layers[0])
        features = [self.layer_block_modules[0](last_layer)]

        feat_shape = list(layers[1].shape[-2:])
        for idx, layer in enumerate(layers[1:], start=1):
            up_sample = F.interpolate(last_layer, size=feat_shape)

            # undergoes a 1x1 convolutional layer to reduce channel dimensions
            inner_layer = self.inner_block_modules[idx](layer)
            last_layer = inner_layer + up_sample
            feat = self.layer_block_modules[idx](last_layer)
            features.append(feat)
            feat_shape[0] *= 2
            feat_shape[1] *= 2

        return features


class ResNet34Backbone(FPN.Backbone):

    def __init__(self, resnet34: ResNet34):
        super().__init__()
        self._resnet34 = resnet34

    @property
    def in_channels_tuple(self) -> tuple:
        return 512, 256, 128, 64

    def bottom_up(self, x: torch.Tensor) -> tuple:
        x = self._resnet34.conv1(x)

        layer1: torch.Tensor = self._resnet34.conv2_x(x)
        layer2: torch.Tensor = self._resnet34.conv3_x(layer1)
        layer3: torch.Tensor = self._resnet34.conv4_x(layer2)
        layer4: torch.Tensor = self._resnet34.conv5_x(layer3)
        return layer4, layer3, layer2, layer1


class MobileNetV3LargeBackbone(FPN.Backbone):

    def __init__(self, mobilenet_v3: MobileNetV3Large):
        super().__init__()
        self._mobilenet_v3 = mobilenet_v3

    @property
    def in_channels_tuple(self) -> tuple:
        return 960, 80, 40, 24, 16

    def bottom_up(self, x: torch.Tensor) -> tuple:
        layer1: torch.Tensor = self._mobilenet_v3.conv1(x)
        layer2: torch.Tensor = self._mobilenet_v3.bneck1(layer1)
        layer3: torch.Tensor = self._mobilenet_v3.bneck2(layer2)
        layer4: torch.Tensor = self._mobilenet_v3.bneck3(layer3)
        layer5: torch.Tensor = self._mobilenet_v3.conv2(self._mobilenet_v3.bneck4(layer4))
        return layer5, layer4, layer3, layer2, layer1



# -*- coding: utf-8 -*-
from typing import List

import torch

from nn_torch.cv.models.fpn import FPN, ResNet34Backbone, MobileNetV3LargeBackbone
from nn_torch.cv.models.mobilenet import MobileNetV3Large
from nn_torch.cv.models.resnet import ResNet34


def test_mobilenet_v3():
    batch_size = 10
    num_class = 4
    net = MobileNetV3Large(num_class)
    x = torch.rand(batch_size, 3, 224, 224)
    y = net(x)
    assert y.shape == torch.Size([batch_size, num_class])


def test_fpn():
    out_channels = 256
    batch_size = 10

    fpns: List[FPN] = [FPN(ResNet34Backbone(ResNet34()), features_num=3, out_channels=out_channels),
                       FPN(MobileNetV3LargeBackbone(MobileNetV3Large()), features_num=3, out_channels=out_channels)]

    x = torch.rand([batch_size, 3, 224, 224])
    for fpn in fpns:
        features = fpn(x)
        assert len(features) == len(fpn.body.in_channels_tuple)
        last_feature_sizes = list(features[0].shape[-2:])
        for feat in features:
            assert feat.shape == torch.Size([batch_size, out_channels] + last_feature_sizes)
            last_feature_sizes[0] *= 2
            last_feature_sizes[1] *= 2
            print(feat.shape)



# -*- coding: utf-8 -*-
"""
残差神经网络模块
"""

import torch
import torch.nn as nn
from torch import Tensor

__all__ = ["ResNet34"]


class ResidualBlock(nn.Module):
    """残差块，用于构造ResNet"""

    def __init__(self, input_channels: int, out_channels: int, *, first_stride=1):
        super().__init__()
        assert first_stride == 1 or first_stride == 2

        self.relu = nn.ReLU(inplace=True)
        self.layer1 = nn.Sequential(
            nn.Conv2d(input_channels, out_channels, kernel_size=3, stride=first_stride, padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )

        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(out_channels)

        self.projection = nn.Conv2d(input_channels, out_channels, kernel_size=1,
                                    stride=2, bias=False) if first_stride == 2 else None

    def forward(self, x):
        identity = x

        out = self.layer1(x)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.projection is not None:
            identity = self.projection(identity)

        out += identity
        return self.relu(out)


class ResNet34(nn.Module):
    """
    残差神经网络
    References:
        https://arxiv.org/abs/1512.03385
    """

    def __init__(self, class_num: int = 1000):
        super().__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3,
                      bias=False),
            nn.BatchNorm2d(64),
        )

        self.conv2_x = nn.Sequential(nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
                                     ResNet34._residual_blocks(64, 3))

        self.conv3_x = nn.Sequential(
            ResidualBlock(64, 128, first_stride=2),
            ResNet34._residual_blocks(128, 3)
        )

        self.conv4_x = nn.Sequential(
            ResidualBlock(128, 256, first_stride=2),
            ResNet34._residual_blocks(256, 5)
        )

        self.conv5_x = nn.Sequential(
            ResidualBlock(256, 512, first_stride=2),
            ResNet34._residual_blocks(512, 2)
        )

        self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512, class_num)

    @staticmethod
    def _residual_blocks(channels: int, layer_num):
        layers = []
        for _ in range(layer_num):
            layers.append(ResidualBlock(channels, channels))

        return nn.Sequential(*layers)

    def forward(self, x: Tensor):
        x = self.conv1(x)
        x = self.conv2_x(x)
        x = self.conv3_x(x)
        x = self.conv4_x(x)
        x = self.conv5_x(x)

        x = self.avg_pool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)
        return x



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

from nn_torch import __version__


def _get_install_requires():
    with open("requirements.txt") as f:
        lines = f.readlines()
        return list(map(lambda s: s.strip(), lines))


setup(
    name="nn_torch",
    version=__version__,
    keywords=("pip", "pytorch", "neural network", "deep learning"),
    description="neural networks implemented by pytorch",
    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=_get_install_requires()
)



# -*- coding: utf-8 -*-
from enum import IntEnum
from typing import Union

import numpy as np
import torch


class BBoxFmt(IntEnum):
    """
    Bound box format enums.
    """
    X_MIN_Y_MIN_WH = 0
    X_MIN_Y_MIN_X_MAX_Y_MAX = 1
    X_C_Y_C_WH = 2


def calc_iou(box1: Union[np.ndarray, torch.Tensor, list],
             box2: Union[np.ndarray, torch.Tensor, list],
             fmt: BBoxFmt):
    """
    compute the iou of two boxes.
    Args:
        box1: bound box1
        box2: bound box2
        fmt: format of bound box
    Return:
        iou: iou of box1 and box2.
    """
    if fmt == BBoxFmt.X_MIN_Y_MIN_X_MAX_Y_MAX:
        x1_min, y1_min, x1_max, y1_max = box1
        x2_min, y2_min, x2_max, y2_max = box2
    elif fmt == BBoxFmt.X_C_Y_C_WH:
        x1_min, y1_min = box1[0] - box1[2] / 2, box1[1] - box1[3] / 2
        x1_max, y1_max = box1[0] + box1[2] / 2, box1[1] + box1[3] / 2
        x2_min, y2_min = box2[0] - box2[2] / 2, box2[1] - box2[3] / 2
        x2_max, y2_max = box2[0] + box2[2] / 2, box2[1] + box2[3] / 2
    elif fmt == BBoxFmt.X_MIN_Y_MIN_WH:
        x1_min, x2_min = box1[0], box2[0]
        y1_min, y2_min = box1[1], box2[1]
        x1_max, x2_max = x1_min + box1[2], x2_min + box2[2]
        y1_max, y2_max = y1_min + box1[3], y2_min + box2[3]
    else:
        raise ValueError('fmt not implemented')

    # calculate intersection rectangle's coordinate
    x_i1 = np.max([x1_min, x2_min])
    y_i1 = np.max([y1_min, y2_min])
    x_i2 = np.min([x1_max, x2_max])
    y_i2 = np.min([y1_max, y2_max])

    # calculate two bound boxes' areas
    area1 = (x1_max - x1_min) * (y1_max - y1_min)
    area2 = (x2_max - x2_min) * (y2_max - y2_min)

    # calculate intersection area
    inter_area = (np.max([0, x_i2 - x_i1])) * (np.max([0, y_i2 - y_i1]))
    iou = inter_area / (area1 + area2 - inter_area + 1e-6)

    return iou



# -*- coding: utf-8 -*-

def smooth_l1(x):
    """
    Smooth L1 loss defined in Fast R-CNN

    References:
        http://arxiv.org/abs/1504.08083

    """
    return 0.5 * x ** 2 if abs(x) < 1 else abs(x) - 0.5



# -*- coding: utf-8 -*-
import abc

import torch.nn as nn
import torch.nn.functional as F
from torch.nn import init


class Hswish(nn.Module):
    def forward(self, x):
        return x * F.relu6(x + 3, inplace=True) / 6


class Hsigmoid(nn.Module):
    def forward(self, x):
        return F.relu6(x + 3, inplace=True) / 6


class SeModule(nn.Module):
    """
    squeeze-and-excite
    """

    def __init__(self, in_size, reduction=4):
        super(SeModule, self).__init__()
        self.se = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_size, in_size // reduction, kernel_size=1, stride=1, padding=0, bias=False),
            nn.BatchNorm2d(in_size // reduction),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_size // reduction, in_size, kernel_size=1, stride=1, padding=0, bias=False),
            nn.BatchNorm2d(in_size),
            Hsigmoid()
        )

    def forward(self, x):
        return x * self.se(x)


class Block(nn.Module):
    """expand + depthwise + pointwise"""

    def __init__(self, kernel_size, in_channels, expand_size, out_channels, nonlinearity, semodule, stride):
        super(Block, self).__init__()
        self.stride = stride
        self.se = semodule

        self.conv1 = nn.Conv2d(in_channels, expand_size, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn1 = nn.BatchNorm2d(expand_size)
        self.nolinear1 = nonlinearity
        self.conv2 = nn.Conv2d(expand_size, expand_size, kernel_size=kernel_size, stride=stride,
                               padding=kernel_size // 2, groups=expand_size, bias=False)
        self.bn2 = nn.BatchNorm2d(expand_size)
        self.nolinear2 = nonlinearity
        self.conv3 = nn.Conv2d(expand_size, out_channels, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn3 = nn.BatchNorm2d(out_channels)

        if stride == 1 and in_channels != out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1, padding=0, bias=False),
                nn.BatchNorm2d(out_channels),
            )
        else:
            self.shortcut = nn.Sequential()

    def forward(self, x):
        out = self.nolinear1(self.bn1(self.conv1(x)))
        out = self.nolinear2(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        if self.se is not None:
            out = self.se(out)
        out = out + self.shortcut(x) if self.stride == 1 else out
        return out


class MobileNetV3(nn.Module, metaclass=abc.ABCMeta):
    """
    References:
        https://arxiv.org/abs/1905.02244
    """

    def __init__(self, num_classes: int):
        super().__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(3, 16, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(16),
            Hswish()
        )

        # initializing bneck conv2 linear3 is in subclass

        self.linear4 = nn.Linear(1280, num_classes)

    def forward(self, x):
        out = self.conv1(x)
        out = self.bneck1(out)
        out = self.bneck2(out)
        out = self.bneck3(out)
        out = self.bneck4(out)
        out = self.conv2(out)
        out = F.avg_pool2d(out, 7)
        out = out.view(out.size(0), -1)
        out = self.linear3(out)
        out = self.linear4(out)
        return out

    def _init_params(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                init.constant_(m.weight, 1)
                init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                init.normal_(m.weight, std=0.001)
                if m.bias is not None:
                    init.constant_(m.bias, 0)


class MobileNetV3Large(MobileNetV3):

    def __init__(self, num_classes: int = 10000):
        super().__init__(num_classes)
        self.bneck1 = nn.Sequential(
            Block(3, 16, 16, 16, nn.ReLU(inplace=True), None, 1),
            Block(3, 16, 64, 24, nn.ReLU(inplace=True), None, 2),
            Block(3, 24, 72, 24, nn.ReLU(inplace=True), None, 1)
        )
        self.bneck2 = nn.Sequential(
            Block(5, 24, 72, 40, nn.ReLU(inplace=True), SeModule(40), 2),
            Block(5, 40, 120, 40, nn.ReLU(inplace=True), SeModule(40), 1),
            Block(5, 40, 120, 40, nn.ReLU(inplace=True), SeModule(40), 1)
        )
        self.bneck3 = nn.Sequential(
            Block(3, 40, 240, 80, Hswish(), None, 2),
            Block(3, 80, 200, 80, Hswish(), None, 1),
            Block(3, 80, 184, 80, Hswish(), None, 1),
            Block(3, 80, 184, 80, Hswish(), None, 1)
        )
        self.bneck4 = nn.Sequential(
            Block(3, 80, 480, 112, Hswish(), SeModule(112), 1),
            Block(3, 112, 672, 112, Hswish(), SeModule(112), 1),
            Block(5, 112, 672, 160, Hswish(), SeModule(160), 1),
            Block(5, 160, 672, 160, Hswish(), SeModule(160), 2),
            Block(5, 160, 960, 160, Hswish(), SeModule(160), 1),
        )

        self.conv2 = nn.Sequential(
            nn.Conv2d(160, 960, kernel_size=1, bias=False),
            nn.BatchNorm2d(960),
            Hswish()
        )
        self.linear3 = nn.Sequential(
            nn.Linear(960, 1280),
            nn.BatchNorm1d(1280),
            Hswish()
        )

        self._init_params()


class MobileNetV3Small(MobileNetV3):

    def __init__(self, num_classes: int = 10000):
        super().__init__(num_classes)
        self.bneck1 = nn.Sequential(
            Block(3, 16, 16, 16, nn.ReLU(inplace=True), SeModule(16), 2),
        )
        self.bneck2 = nn.Sequential(
            Block(3, 16, 72, 24, nn.ReLU(inplace=True), None, 2),
            Block(3, 24, 88, 24, nn.ReLU(inplace=True), None, 1)
        )
        self.bneck3 = nn.Sequential(
            Block(5, 24, 96, 40, Hswish(), SeModule(40), 2),
            Block(5, 40, 240, 40, Hswish(), SeModule(40), 1),
            Block(5, 40, 240, 40, Hswish(), SeModule(40), 1),
            Block(5, 40, 120, 48, Hswish(), SeModule(48), 1),
            Block(5, 48, 144, 48, Hswish(), SeModule(48), 1)
        )
        self.bneck4 = nn.Sequential(
            Block(5, 48, 288, 96, Hswish(), SeModule(96), 2),
            Block(5, 96, 576, 96, Hswish(), SeModule(96), 1),
            Block(5, 96, 576, 96, Hswish(), SeModule(96), 1),
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(96, 576, kernel_size=1, bias=False),
            nn.BatchNorm2d(576),
            Hswish()
        )
        self.linear3 = nn.Sequential(
            nn.Linear(576, 1280),
            nn.BatchNorm1d(1280),
            Hswish()
        )

        self._init_params()
